from ailment.expression import Convert, BinaryOp, Const, UnaryOp

from .base import PeepholeOptimizationExprBase


class OneSubBool(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "1 - bool_expr => !bool_expr"
    expr_classes = (BinaryOp, )  # all expressions are allowed

    def optimize(self, expr: BinaryOp):

        # Sub(1, Conv(1->N, some bool expression)) ==> Conv(1->N, Not(some bool expression))
        if expr.op == "Sub" and isinstance(expr.operands[0], Const) and expr.operands[0].value == 1 \
                and isinstance(expr.operands[1], Convert) and expr.operands[1].from_bits == 1:
            conv_expr = expr.operands[1]
            if self.is_bool_expr(conv_expr.operand):
                new_expr = Convert(None, 1, conv_expr.to_bits, conv_expr.is_signed,
                                   UnaryOp(None, 'Not', conv_expr.operand, **conv_expr.operand.tags),
                                   **conv_expr.tags,
                                   )
                return new_expr

        return None



from ailment.expression import BinaryOp, BasePointerOffset, Const

from .base import PeepholeOptimizationExprBase


class BasePointerOffsetAddN(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "(Ptr - M) + N => Ptr - (M - N)"
    expr_classes = (BinaryOp, )  # all expressions are allowed

    def optimize(self, expr: BinaryOp):

        if (expr.op in ("Add", "Sub")
                and isinstance(expr.operands[0], BasePointerOffset)
                and isinstance(expr.operands[1], Const)
        ):
            offset = expr.operands[0].offset
            if expr.op == "Add":
                offset += expr.operands[1].value
            else:  # expr.op == "Sub"
                offset -= expr.operands[1].value

            # convert offset to a signed integer
            max_int = (1 << (self.project.arch.bits - 1)) - 1
            if offset > max_int:
                offset -= 1 << self.project.arch.bits

            r = expr.operands[0].copy()
            r.offset = offset
            return r

        return None



from . import Analysis
from .. import SIM_PROCEDURES

import logging
l = logging.getLogger(name=__name__)

class CalleeCleanupFinder(Analysis):
    def __init__(self, starts=None, hook_all=False):
        self.results = {}

        if starts is None:
            starts = [imp.resolvedby.rebased_addr for imp in self.project.loader.main_object.imports.values()]

        for addr in starts:
            with self._resilience():
                size = self.analyze(addr)
                if size is None:
                    l.info("Couldn't find return for function at %#x", addr)
                else:
                    self.results[addr] = size

        if hook_all:
            for addr, size in self.results.items():
                if self.project.is_hooked(addr):
                    continue
                if size % self.project.arch.bytes != 0:
                    l.error("Function at %#x has a misaligned return?", addr)
                    continue
                args = size // self.project.arch.bytes
                cc = self.project.factory.cc()
                prototype = cc.guess_prototype([0]*args)
                cc.CALLEE_CLEANUP = True
                sym = self.project.loader.find_symbol(addr)
                name = sym.name if sym is not None else None
                lib = self.project.loader.find_object_containing(addr)
                libname = lib.provides if lib is not None else None
                self.project.hook(addr, SIM_PROCEDURES['stubs']['ReturnUnconstrained'](cc=cc, prototype=prototype, display_name=name, library_name=libname, is_stub=True))

    def analyze(self, addr):
        seen = set()
        todo = [addr]

        while todo:
            addr = todo.pop(0)
            seen.add(addr)
            irsb = self.project.factory.block(addr, opt_level=0).vex
            if irsb.jumpkind == 'Ijk_Ret':
                # got it!
                for stmt in reversed(irsb.statements):
                    if stmt.tag == 'Ist_IMark':
                        l.error("VERY strange return instruction at %#x...", addr)
                        break
                    if stmt.tag == 'Ist_WrTmp':
                        if stmt.data.tag == 'Iex_Binop':
                            if stmt.data.op.startswith('Iop_Add'):
                                return stmt.data.args[1].con.value - self.project.arch.bytes
            elif irsb.jumpkind == 'Ijk_Call':
                if addr + irsb.size not in seen:
                    todo.append(addr + irsb.size)
            else:
                todo.extend(irsb.constant_jump_targets - seen)

        return None

from angr.analyses import AnalysesHub
AnalysesHub.register_default('CalleeCleanupFinder', CalleeCleanupFinder)



# pylint:disable=too-many-boolean-expressions
import logging
from typing import Optional, Union, Type, Iterable, Tuple, Set, TYPE_CHECKING

from ailment.statement import Statement, Assignment, Call, Store, Jump
from ailment.expression import Expression, Tmp, Load, Const

from ...engines.light.data import SpOffset
from ...knowledge_plugins.key_definitions.constants import OP_AFTER
from ...knowledge_plugins.key_definitions import atoms
from ...analyses.reaching_definitions.external_codeloc import ExternalCodeLocation
from ...analyses.propagator import PropagatorAnalysis
from ...analyses.reaching_definitions import ReachingDefinitionsAnalysis
from ...errors import SimMemoryMissingError
from .. import Analysis, register_analysis
from .peephole_optimizations import STMT_OPTS, EXPR_OPTS, PeepholeOptimizationStmtBase, PeepholeOptimizationExprBase
from .ailblock_walker import AILBlockWalker

if TYPE_CHECKING:
    from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
    from angr.knowledge_plugins.key_definitions.live_definitions import LiveDefinitions, Definition
    from ailment.block import Block


_l = logging.getLogger(name=__name__)


class BlockSimplifier(Analysis):
    """
    Simplify an AIL block.
    """
    def __init__(self, block: Optional['Block'], func_addr: Optional[int] = None,
                 remove_dead_memdefs=False, stack_pointer_tracker=None,
                 peephole_optimizations: Optional[Iterable[Union[Type[PeepholeOptimizationStmtBase],
                                                                 Type[PeepholeOptimizationExprBase]]]]=None,
                 stack_arg_offsets: Optional[Set[Tuple[int, int]]] = None,
                 cached_reaching_definitions=None,
                 cached_propagator=None,
                 ):
        """
        :param block:   The AIL block to simplify. Setting it to None to skip calling self._analyze(), which is useful
                        in test cases.
        """

        self.block = block
        self.func_addr = func_addr

        self._remove_dead_memdefs = remove_dead_memdefs
        self._stack_arg_offsets = stack_arg_offsets
        self._stack_pointer_tracker = stack_pointer_tracker

        if peephole_optimizations is None:
            self._expr_peephole_opts = [ cls(self.project, self.kb, self.func_addr) for cls in EXPR_OPTS ]
            self._stmt_peephole_opts = [ cls(self.project, self.kb, self.func_addr) for cls in STMT_OPTS ]
        else:
            self._expr_peephole_opts = [ cls(self.project, self.kb, self.func_addr) for cls in peephole_optimizations
                                         if issubclass(cls, PeepholeOptimizationExprBase) ]
            self._stmt_peephole_opts = [ cls(self.project, self.kb, self.func_addr) for cls in peephole_optimizations
                                         if issubclass(cls, PeepholeOptimizationStmtBase) ]

        self.result_block = None

        # cached Propagator and ReachingDefinitions results. Clear them if the block is updated
        self._propagator = cached_propagator
        self._reaching_definitions = cached_reaching_definitions

        if self.block is not None:
            self._analyze()

    def _analyze(self):

        block = self.block
        ctr = 0
        max_ctr = 30

        new_block = self._eliminate_self_assignments(block)
        if self._count_nonconstant_statements(new_block) >= 2 and self._has_propagatable_assignments(new_block):
            new_block = self._eliminate_dead_assignments(new_block)
        if new_block != block:
            self._clear_cache()
            block = new_block

        while True:
            ctr += 1
            # print(str(block))
            new_block = self._simplify_block_once(block)
            # print()
            # print(str(new_block))
            if new_block == block:
                break
            self._clear_cache()
            block = new_block
            if ctr >= max_ctr:
                _l.error("Simplification does not reach a fixed point after %d iterations. "
                         "Block comparison is probably incorrect.", max_ctr)
                break

        self.result_block = block

    def _compute_propagation(self, block):
        if self._propagator is None:
            self._propagator = self.project.analyses[PropagatorAnalysis].prep()(
                block=block, func_addr=self.func_addr, stack_pointer_tracker=self._stack_pointer_tracker,
            )
        return self._propagator

    def _compute_reaching_definitions(self, block):
        if self._reaching_definitions is None:
            self._reaching_definitions = self.project.analyses[ReachingDefinitionsAnalysis].prep()(
                subject=block, track_tmps=True, observation_points=[('node', block.addr, OP_AFTER)]
            )
        return self._reaching_definitions

    def _clear_cache(self):
        self._reaching_definitions = None
        self._propagator = None

    @staticmethod
    def _has_propagatable_assignments(block) -> bool:
        return any(isinstance(stmt, (Assignment, Store)) for stmt in block.statements)

    @staticmethod
    def _count_nonconstant_statements(block) -> int:
        return sum(1 for stmt in block.statements if not (isinstance(stmt, Jump) and isinstance(stmt.target, Const)))

    def _simplify_block_once(self, block):
        nonconstant_stmts = self._count_nonconstant_statements(block)
        has_propagatable_assignments = self._has_propagatable_assignments(block)

        # propagator
        if nonconstant_stmts >= 2 and has_propagatable_assignments:
            propagator = self._compute_propagation(block)
            replacements = list(propagator._states.values())[0]._replacements
            if replacements:
                _, new_block = self._replace_and_build(block, replacements)
                new_block = self._eliminate_self_assignments(new_block)
                self._clear_cache()
            else:
                new_block = block
        else:
            # Skipped calling Propagator
            new_block = block

        if nonconstant_stmts >= 2 and has_propagatable_assignments:
            new_block = self._eliminate_dead_assignments(new_block)

        new_block = self._peephole_optimize(new_block)
        return new_block

    @staticmethod
    def _replace_and_build(block, replacements, replace_assignment_dsts: bool=False,
                           gp: Optional[int]=None) -> Tuple[bool,'Block']:

        new_statements = block.statements[::]
        replaced = False

        for codeloc, repls in replacements.items():
            for old, new in repls.items():
                stmt = new_statements[codeloc.stmt_idx]
                if isinstance(old, Load) \
                        and not isinstance(stmt, Call) \
                        and not (gp is not None and isinstance(new, Const) and new.value == gp):
                    # skip memory-based replacement for non-Call and non-gp-loading statements
                    continue
                if stmt == old:
                    # replace this statement
                    r = True
                    new_stmt = new
                else:
                    # replace the expressions involved in this statement
                    if isinstance(stmt, Call) and isinstance(new, Call) and old == stmt.ret_expr:
                        # special case: do not replace the ret_expr of a call statement to another call statement
                        r = False
                        new_stmt = None
                    elif isinstance(stmt, Assignment) and not replace_assignment_dsts:
                        # special case: do not replace the dst
                        new_stmt = None
                        if stmt.src == old:
                            r = True
                            new_src = new.copy()
                        else:
                            r, new_src = stmt.src.replace(old, new)
                        if r:
                            new_stmt = Assignment(stmt.idx, stmt.dst, new_src, **stmt.tags)
                    else:
                        r, new_stmt = stmt.replace(old, new)

                if r:
                    replaced = True
                    new_statements[codeloc.stmt_idx] = new_stmt

        if not replaced:
            return False, block

        new_block = block.copy()
        new_block.statements = new_statements
        return True, new_block

    @staticmethod
    def _eliminate_self_assignments(block):

        new_statements = [ ]

        for stmt in block.statements:
            if type(stmt) is Assignment:
                if stmt.dst.likes(stmt.src):
                    continue
            new_statements.append(stmt)

        new_block = block.copy(statements=new_statements)
        return new_block

    def _eliminate_dead_assignments(self, block):

        new_statements = [ ]
        if not block.statements:
            return block

        rd = self._compute_reaching_definitions(block)
        live_defs: 'LiveDefinitions' = rd.one_result

        # Find dead assignments
        dead_defs_stmt_idx = set()
        all_defs: Iterable['Definition'] = rd.all_definitions
        stackarg_offsets = set(tpl[1] for tpl in self._stack_arg_offsets) \
            if self._stack_arg_offsets is not None else None
        for d in all_defs:
            if isinstance(d.codeloc, ExternalCodeLocation) or d.dummy:
                continue
            if isinstance(d.atom, atoms.MemoryLocation):
                if not self._remove_dead_memdefs:
                    # we always remove definitions for stack arguments
                    if stackarg_offsets is not None and isinstance(d.atom.addr, atoms.SpOffset):
                        if d.atom.addr.offset not in stackarg_offsets:
                            continue
                    else:
                        continue

            if isinstance(d.atom, atoms.Tmp):
                uses = live_defs.tmp_uses[d.atom.tmp_idx]
                if not uses:
                    dead_defs_stmt_idx.add(d.codeloc.stmt_idx)
            else:
                uses = rd.all_uses.get_uses(d)
                if not uses:
                    # it's entirely possible that at the end of the block, a register definition is not used.
                    # however, it might be used in future blocks.
                    # so we only remove a definition if the definition is not alive anymore at the end of the block
                    defs_ = set()
                    if isinstance(d.atom, atoms.Register):
                        try:
                            vs: 'MultiValues' = live_defs.register_definitions.load(d.atom.reg_offset, size=d.atom.size)
                        except SimMemoryMissingError:
                            vs = None
                    elif isinstance(d.atom, atoms.MemoryLocation) and isinstance(d.atom.addr, SpOffset):
                        stack_addr = live_defs.stack_offset_to_stack_addr(d.atom.addr.offset)
                        try:
                            vs: 'MultiValues' = live_defs.stack_definitions.load(stack_addr, size=d.atom.size,
                                                                                 endness=d.atom.endness)
                        except SimMemoryMissingError:
                            vs = None
                    else:
                        continue
                    if vs is not None:
                        for values in vs.values():
                            for value in values:
                                defs_.update(live_defs.extract_defs(value))

                    if d not in defs_:
                        dead_defs_stmt_idx.add(d.codeloc.stmt_idx)

        used_tmps = set()
        # micro optimization: if all statements that use a tmp are going to be removed, we remove this tmp as well
        for tmp, used_locs in rd.one_result.tmp_uses.items():
            used_at = set(loc.stmt_idx for loc in used_locs)
            if used_at.issubset(dead_defs_stmt_idx):
                continue
            used_tmps.add(tmp)

        # Remove dead assignments
        for idx, stmt in enumerate(block.statements):
            if type(stmt) is Assignment:
                if type(stmt.dst) is Tmp:
                    if stmt.dst.tmp_idx not in used_tmps:
                        continue

                # is it a dead virgin?
                if idx in dead_defs_stmt_idx:
                    continue

                if stmt.src == stmt.dst:
                    continue

            new_statements.append(stmt)

        new_block = block.copy(statements=new_statements)
        return new_block

    #
    # Peephole optimization
    #

    def _peephole_optimize(self, block):

        # expressions are updated in place
        self._peephole_optimize_exprs(block, self._expr_peephole_opts)

        statements, stmts_updated = self._peephole_optimize_stmts(block, self._stmt_peephole_opts)

        if not stmts_updated:
            return block
        new_block = block.copy(statements=statements)
        return new_block

    @staticmethod
    def _peephole_optimize_exprs(block, expr_opts):

        class _any_update:
            v = False

        def _handle_expr(expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement, block) -> Optional[Expression]:

            old_expr = expr

            redo = True
            while redo:
                redo = False
                for expr_opt in expr_opts:
                    if isinstance(expr, expr_opt.expr_classes):
                        r = expr_opt.optimize(expr)
                        if r is not None and r is not expr:
                            expr = r
                            redo = True
                            break

            if expr is not old_expr:
                _any_update.v = True
                # continue to process the expr
                r = AILBlockWalker._handle_expr(walker, expr_idx, expr, stmt_idx, stmt, block)
                return expr if r is None else r

            return AILBlockWalker._handle_expr(walker, expr_idx, expr, stmt_idx, stmt, block)

        # run expression optimizers
        walker = AILBlockWalker()
        walker._handle_expr = _handle_expr
        walker.walk(block)

        return _any_update.v

    @staticmethod
    def _peephole_optimize_stmts(block, stmt_opts):

        any_update = False
        statements = [ ]

        # run statement optimizers
        for stmt in block.statements:
            old_stmt = stmt
            redo = True
            while redo:
                redo = False
                for opt in stmt_opts:
                    if isinstance(stmt, opt.stmt_classes):
                        r = opt.optimize(stmt)
                        if r is not None and r is not stmt:
                            stmt = r
                            redo = True
                            break

            if stmt is not None and stmt is not old_stmt:
                statements.append(stmt)
                any_update = True
            else:
                statements.append(old_stmt)

        return statements, any_update


register_analysis(BlockSimplifier, 'AILBlockSimplifier')




from archinfo.arch_soot import SootArgument, SootMethodDescriptor

from . import translate_expr
from ..method_dispatcher import resolve_method
from ..exceptions import SootMethodNotLoadedException
from .base import SimSootExpr


class InvokeBase(SimSootExpr):
    def __init__(self, expr, state):
        super(InvokeBase, self).__init__(expr, state)
        self.method = None
        self.args = None

    def _execute(self):
        # the invoke target gets resolved differently based on the invoke type
        # => static vs. virtual vs. special (= non-virtual)
        self.method = self._resolve_invoke_target(self.expr, self.state)
        # translate arguments
        self.args = self._translate_args()

    def _translate_args(self):
        args = []
        # for instance method calls, add the 'this' reference
        if isinstance(self, SimSootExpr_VirtualInvoke) or \
           isinstance(self, SimSootExpr_SpecialInvoke):
            this_ref_base = self._translate_value(self.expr.base)
            this_ref = self.state.memory.load(this_ref_base, none_if_missing=True)
            this_ref_type = this_ref.type if this_ref is not None else None
            args += [SootArgument(this_ref, this_ref_type, is_this_ref=True)]

        # translate function arguments
        for arg in self.expr.args:
            if "Constant" in arg.__class__.__name__:
                # argument is a constant
                # => translate the expr to get the value
                arg_value = self._translate_expr(arg).expr
            else:
                # argument is a variable
                # => load value from memory
                arg_value = self.state.memory.load(self._translate_value(arg),
                                                   none_if_missing=True)
            args += [SootArgument(arg_value, arg.type)]

        return args

    def _resolve_invoke_target(self, expr, state):
        raise NotImplementedError()


class SimSootExpr_VirtualInvoke(InvokeBase):
    """
    Instance methods are dynamically resolved by the actual type of the
    base object (i.e. the object on which the method is invoked) and *not*
    by the type of the variable storing the reference.
    """
    def _resolve_invoke_target(self, expr, state):
        # get the type of the base object
        base = translate_expr(self.expr.base, self.state).expr
        if base is not None:
            base_type = base.type
        else:
            # the base is not set, for example if we process an invocation of an
            # unloaded library function
            # => fallback: use the statically retrieved type
            base_type = self.expr.class_name

        # based on the class of the base object, we resolve the invoke target
        try:
            return resolve_method(state=self.state,
                                  method_name=self.expr.method_name,
                                  class_name=base_type,
                                  params=self.expr.method_params,
                                  ret_type=self.expr.type,
                                  raise_exception_if_not_found=True)
        except SootMethodNotLoadedException:
            # in case that the method is not loaded, continue with the infos
            # available from the invoke expression
            return SootMethodDescriptor(self.expr.class_name,
                                        self.expr.method_name,
                                        self.expr.method_params,
                                        ret_type=self.expr.type)



class SimSootExpr_SpecialInvoke(InvokeBase):
    """
    Special invocations are used either for invoking instance methods of a
    superclass or a constructor. Compared to virtual invokes, the class
    containing the method is passed explicitly in the invoke expression
    (@expr.class_name) rather than determined dynamically by the type of the
    base objects.
    """
    def _resolve_invoke_target(self, expr, state):
        return resolve_method(state=self.state,
                              method_name=self.expr.method_name,
                              class_name=self.expr.class_name,
                              params=self.expr.method_params,
                              ret_type=self.expr.type)


class SimSootExpr_StaticInvoke(InvokeBase):
    def _resolve_invoke_target(self, expr, state):
        return resolve_method(state=self.state,
                              method_name=self.expr.method_name,
                              class_name=self.expr.class_name,
                              params=self.expr.method_params,
                              ret_type=self.expr.type)

class SimSootExpr_InterfaceInvoke(SimSootExpr_VirtualInvoke):
    def _resolve_invoke_target(self, expr, state):
        return super(SimSootExpr_InterfaceInvoke, self)._resolve_invoke_target(expr, state)




import logging

from .base import SimSootStmt

l = logging.getLogger('angr.engines.soot.statements.invoke')


class SimSootStmt_Invoke(SimSootStmt):
    def _execute(self):
        invoke_expr = self._translate_expr(self.stmt.invoke_expr)
        self._add_invoke_target(invoke_expr)
